#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    The MIT License (MIT)

    Copyright (c) 2015+ Raphaël SEBAN <motus@laposte.net>

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use, copy,
    modify, merge, publish, distribute, sublicense, and/or sell copies
    of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
    BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
    ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
    CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

import os
import queue
import threading
import time
import tkinter as tk
from tkinter import ttk
import urllib.parse
import urllib.request


def _ (text):
    """Looking for i18n translation support.

    This technique allows the use of internationalization (i18n)
    parsers such as GNU xgettext, by marking each string to be
    translated with _("text to translate").

    This also makes this module compatible with tkRAD i18n support,
    with nothing else than merely importing it and using it straight
    forward."""
    try:
        return __builtins__["_"](text)
    except:
        return text
    # end try
# end def


def get_downloader (**kw):
    return DownloaderController(**kw)
# end def


def shorten_path (path, width=40, placeholder="..."):
    if not placeholder:
        raise ValueError(
            _("A placeholder string of chars is required.")
        )
    else:
        width = max(0, width)
        if len(path) > width:
            (_head, _tail) = os.path.split(path)
            placeholder = placeholder[:width]
            _swidth = width - len(placeholder)
            if _swidth > 0:
                _tail = _tail[-_swidth:]
                _swidth = max(0, _swidth - len(_tail))
                _head = _head[:_swidth]
            else:
                _head = _tail = ""
            # end if
            path = "{}{}{}".format(_head, placeholder, _tail)
        # end if
    # end if
    return path
# end def


def tk_get_downloaded_filepath (url, download_directory):
    try:
        tk_parent = tk._default_root.winfo_toplevel()
    except:
        tk_parent = None
    # end try
    return tkDownloaderDialog(tk_parent).get_downloaded_filepath(
        url, download_directory
    )
# end def


class DownloaderController:

    def __init__ (self, url=None, download_directory=None,
                                                update_callback=None):
        self.model = DownloaderModel(
            url=url,
            download_directory=download_directory,
            update_callback=update_callback,
        )
        self._started = False
    # end def

    def cancel_download (self, timeout=None):
        return self.model.cancel(timeout)
    # end def

    @property
    def download_directory (self):
        return self.model.download_directory
    # end def

    @download_directory.setter
    def download_directory (self, dir_path):
        self.model.download_directory = dir_path
    # end def

    def get_downloaded_filepath (self, timeout=None):
        return self.model.get_downloaded_filepath(timeout)
    # end def

    def is_downloading (self):
        return self.model.is_alive()
    # end def

    def start_download (self, url=None, download_directory=None,
                                                update_callback=None):
        if self.is_downloading():
            raise DownloaderError(_("Downloader is already running."))
            return
        elif self._started:
            self.model = DownloaderModel(
                url=self.url,
                download_directory=self.download_directory,
                update_callback=self.update_callback,
            )
        # end if
        self.url = url or self.url
        self.download_directory = (
            download_directory or self.download_directory
        )
        self.update_callback = update_callback or self.update_callback
        self.model.start()
        self._started = True
    # end def

    @property
    def update_callback (self):
        return self.model.update_callback
    # end def

    @update_callback.setter
    def update_callback (self, callback):
        self.model.update_callback = callback
    # end def

    @property
    def url (self):
        return self.model.url
    # end def

    @url.setter
    def url (self, url):
        self.model.url = url
    # end def

# end class


class DownloaderError (Exception):
    pass
# end class


class DownloaderMiningDude (threading.Thread):

    def __init__ (self, queue, update_callback):
        super().__init__()
        self.daemon = True
        self.queue = queue
        self.update_callback = update_callback
    # end def

    def run (self):
        _locked = False
        while not (_locked and self.queue.empty()):
            task = self.queue.get()
            data = task.data
            self.queue.task_done()
            if data is None:
                _locked = True
            elif not _locked:
                self.update_callback(**data)
            # end if
        # end while
    # end def

    @property
    def update_callback (self):
        return self.__update_callback
    # end def

    @update_callback.setter
    def update_callback (self, callback):
        if callable(callback):
            self.__update_callback = callback
        else:
            self.__update_callback = lambda *args, **kw: None
        # end if
    # end def

# end class


class DownloaderModel (threading.Thread):

    def __init__ (self, url=None, download_directory=None,
                                                update_callback=None):
        super().__init__()
        self.daemon = True
        self._running = False
        self._cancelled = False
        self._states = dict()
        self._block_size = 8192
        self._file_path = ""
        self._queue = queue.PriorityQueue()
        self.url = url
        self.download_directory = download_directory
        self.update_callback = update_callback
    # end def

    def _do_step_01 (self):
        self._file_path = ""
        _url = shorten_path(self.url)
        self._vars.update(short_url=_url)
        self._update_interface(
            url=self.url,
            shorten_url=_url,
            status="URL_OPEN",
            info=_("Trying to connect to server..."),
        )
        _attempt = self._vars.get("count", 0)
        if _attempt < 3:
            _response = urllib.request.urlopen(self.url)
            _size = _response.getheader("Content-Length")
            if _size:
                self._vars.update(
                    count=0,
                    response=_response,
                    content_length=int(_size),
                )
                return 2
            # end if
            self._vars.update(count=int(_attempt + 1))
            time.sleep(1.0)
            return 1
        else:
            raise DownloaderError(
                _("Unable to determine remote file size.")
            )
            self._running = False
            self._cancelled = True
            return
        # end if
    # end def

    def _do_step_02 (self):
        _ddir = self._get_available_directory(self.download_directory)
        _filename = self._get_available_filename(self.url, _ddir)
        _fpath = os.path.join(_ddir, _filename)
        _ftemp_path = _fpath + ".part"
        _fd = open(_ftemp_path, "wb")
        self._vars.update(
            available_dir=_ddir,
            available_filename=_filename,
            available_filepath=_fpath,
            available_fd=_fd,
            temporary_filepath=_ftemp_path,
        )
        return 3
    # end def

    def _do_step_03 (self):
        _count = self._vars.get("count", 0)
        _fd = self._vars.get("available_fd")
        _response = self._vars.get("response")
        _size = self._vars.get("content_length")
        _url = self._vars.get("short_url")
        self._info("DOWNLOAD", _("Downloading file..."))
        _data = _response.read(self._block_size)
        _fd.write(_data)
        _count += len(_data)
        self._vars.update(count=_count)
        self._progress(100 * _count // _size)
        if _count >= _size:
            _response.close()
            _fd.close()
            os.rename(
                self._vars.get("temporary_filepath"),
                self._vars.get("available_filepath")
            )
            self._file_path = self._vars.get("available_filepath")
            self._running = False
        # end if
        return 3 # loop on step 3
    # end def

    def _get_available_directory (self, dir_list):
        if dir_list:
            if isinstance(dir_list, str):
                dir_list = dir_list.split()
            # end if
            if isinstance(dir_list, (tuple, list)):
                for _dir in dir_list:
                    # ensure directory path is correct
                    _dir = os.path.abspath(
                        os.path.realpath(os.path.expanduser(_dir))
                    )
                    # directory must exist AND being writable
                    if os.path.isdir(_dir) and os.access(_dir, os.W_OK):
                        return _dir
                    # end if
                # end for
            else:
                raise DownloaderError(
                    _("Download directory must either be a "
                        "string of chars or a list of strings.")
                )
                return ""
            # end if
        # end if
        raise DownloaderError(_("No available download directory."))
        return ""
    # end def

    def _get_available_filename (self, url, target_dir):
        _fpath = urllib.parse.urlsplit(url).path
        (_head, _tail) = os.path.split(_fpath)
        (_prefix, _suffix) = os.path.splitext(_tail)
        _prefix = _prefix or "downloaded_file"
        _index = _sep = ""
        while True:
            _fname = "{}{}{}{}".format(_prefix, _sep, _index, _suffix)
            _fpath = os.path.join(target_dir, _fname)
            if not os.path.lexists(_fpath):
                return _fname
            # end if
            if not _index:
                _index = 0
                _sep = "-"
            # end if
            _index += 1
            if _index > 10000:
                break
            # end if
        # end while
        raise DownloaderError(_("No available filename."))
        return ""
    # end def

    def _info (self, state, message, priority=99):
        self._update_interface(
            status=str(state).upper(), info=message, priority=priority
        )
    # end def

    def _progress (self, value, priority=99):
        self._update_interface(progress=value, priority=priority)
    # end def

    def _purge_temporary_file (self):
        try:
            self._vars.get("available_fd").close()
        except:
            pass
        # end try
        try:
            os.remove(self._vars.get("temporary_filepath"))
        except:
            pass
        # end try
    # end def

    def _queue_put (self, data, priority=99):
        self._queue.put(DownloaderPriorityQueueTask(priority, data))
    # end def

    def _release_dude (self, priority=0):
        self._queue_put(None, priority)
    # end def

    def _update_interface (self, priority=99, **kw):
        self._states.update(kw)
        priority += int(kw.get("progress", 0))
        self._queue_put(self._states.copy(), priority)
    # end def

    def cancel (self, timeout=None):
        if self.is_alive():
            self._running = False
            self._cancelled = True
            self.join(timeout)
        # end if
        return not self.is_alive()
    # end def

    def get_downloaded_filepath (self, timeout=None):
        if self.is_alive():
            self.join(timeout)
        # end if
        return self._file_path
    # end def

    def run (self):
        mining_dude = DownloaderMiningDude(
            queue=self._queue, update_callback=self.update_callback
        )
        mining_dude.start()
        self._running = True
        self._cancelled = False
        self._vars = dict()
        _step = 1
        while self._running and not self._cancelled:
            _step = getattr(self, "_do_step_{:02d}".format(_step))()
        # end while
        self._purge_temporary_file()
        if self._cancelled:
            self._info("EXIT_CANCEL", _("Cancelled."), priority=1)
        else:
            self._info("EXIT_SUCCESS", _("Done."), priority=1)
        # end if
        self._release_dude(priority=2)
    # end def

# end class


class DownloaderPriorityQueueTask:

    def __init__ (self, priority=99, data=None):
        self.priority = int(priority)
        self.data = data
    # end def

    def __lt__ (self, right_operand):
        try:
            return int(self.priority) < int(right_operand.priority)
        except:
            return NotImplemented
        # end try
    # end def

# end class


class TkinterDemo (tk.Tk):

    URL = (
        "https://github.com/tarball69/tkBoulderDash/archive/master.zip"
    )

    DOWNLOAD_DIRS = (
        "~/tmp ~/temp ~/download ~/Download "
        "~/downloads ~/Downloads ~ /tmp"
    )

    def __init__ (self):
        super().__init__()
        self.setup_gui()
    # end def

    def download_from_dialog (self):
        self.status_reset()
        _file_path = tk_get_downloaded_filepath(
            self.URL, self.DOWNLOAD_DIRS
        )
        if _file_path:
            self.status_success(_file_path)
        else:
            self.status_cancel()
        # end if
    # end def

    def download_in_frame (self):
        self.status_reset()
        self.dnlframe.download(
            self.URL, self.DOWNLOAD_DIRS,
            on_success=self.on_download_success,
            on_cancel=self.on_download_cancel,
        )
    # end def

    def on_download_cancel (self, *args, **kw):
        self.status_cancel()
    # end def

    def on_download_success (self, file_path, *args, **kw):
        self.status_success(file_path)
    # end def

    def on_quit_demo (self, *args, **kw):
        self.dnlframe.cancel_download()
        self.destroy()
    # end def

    def run (self):
        self.deiconify()
        self.mainloop()
    # end def

    def setup_gui (self):
        self.title(_("Tkinter demo"))
        self.withdraw()
        self.resizable(width=False, height=False)
        self.protocol("WM_DELETE_WINDOW", self.on_quit_demo)
        _frame = ttk.Frame(self, padding=5)
        ttk.Label(
            _frame, text=_("Click the following button to ")
        ).pack(side=tk.LEFT)
        ttk.Button(
            _frame,
            text=_("download in frame"),
            command=self.download_in_frame,
        ).pack(side=tk.RIGHT)
        _frame.pack(anchor=tk.W)
        self.dnlframe = ttkDownloaderFrame(
            self, downloader=None, padding=5
        )
        self.dnlframe.pack(expand=1, fill=tk.BOTH)
        _frame = ttk.Frame(self, padding=5)
        ttk.Label(
            _frame, text=_("Click the following button to ")
        ).pack(side=tk.LEFT)
        ttk.Button(
            _frame,
            text=_("download from a dialog box"),
            command=self.download_from_dialog,
        ).pack(side=tk.RIGHT)
        _frame.pack(anchor=tk.W)
        _frame = ttk.Frame(self, padding=5)
        ttk.Label(_frame, text=_("Status: ")).pack(side=tk.LEFT)
        self.status_cvar = tk.StringVar()
        ttk.Label(
            _frame, textvariable=self.status_cvar, foreground="red",
        ).pack(side=tk.RIGHT, expand=1, fill=tk.X)
        _frame.pack(anchor=tk.W)
        _frame = ttk.Frame(self, padding=5)
        ttk.Button(
            _frame, text=_("Quit demo"), command=self.on_quit_demo,
        ).pack()
        _frame.pack()
    # end def

    def status_cancel (self, reason=""):
        self.status_cvar.set(
            "\n".join((_("download cancelled."), str(reason))).strip()
        )
    # end def

    def status_reset (self):
        self.status_cvar.set("")
    # end def

    def status_success (self, file_path):
        self.status_cvar.set(
            "\n".join((
                _("downloaded file is available at:"), str(file_path)
            )).strip()
        )
    # end def

# end class


class tkDownloaderDialog (tk.Toplevel):

    def __init__ (self, master=None, **options):
        super().__init__(master, **options)
        self.__downloaded_filepath = ""
        self._setup_widget()
    # end def

    def _center_window (self):
        self.update_idletasks()
        if self.master:
            _root = self.master.winfo_toplevel()
            _rw, _rh = _root.winfo_width(), _root.winfo_height()
            _x0, _y0 = _root.winfo_x(), _root.winfo_y()
        else:
            _rw = self.winfo_screenwidth()
            _rh = self.winfo_screenheight()
            _x0 = _y0 = 0
        # end if
        _w, _h = self.winfo_reqwidth(), self.winfo_reqheight()
        self.geometry(
            "+{x}+{y}".format(x=_x0 + (_rw-_w)//2, y=_y0 + (_rh-_h)//2)
        )
    # end def

    def _close_dialog (self):
        self.update_idletasks()
        self.after(500, self.destroy)
    # end def

    def _on_download_cancel (self, *args, **kw):
        self.__downloaded_filepath = ""
        self._close_dialog()
    # end def

    def _on_download_success (self, file_path, *args, **kw):
        self.__downloaded_filepath = file_path
        self._close_dialog()
    # end def

    def _setup_widget (self):
        self.withdraw()
        self.transient(self.master)
        self.resizable(width=False, height=False)
        self.frame = ttkDownloaderFrame(self, padding=10)
        self.frame.pack()
        self.protocol("WM_DELETE_WINDOW", self.on_quit_dialog)
    # end def

    def download (self, url, download_directory):
        self.frame.download(
            url=url,
            download_directory=download_directory,
            on_success=self._on_download_success,
            on_cancel=self._on_download_cancel,
        )
    # end def

    def get_downloaded_filepath (self, url, download_directory):
        self.download(url, download_directory)
        self.show()
        self.grab_set()
        self.wait_window(self)
        return self.__downloaded_filepath
    # end def

    def on_quit_dialog (self, *args, **kw):
        self.frame._on_download_cancel()
    # end def

    def show (self, *args, **kw):
        self.deiconify()
        self._center_window()
    # end def

# end class


class ttkDownloaderFrame (ttk.Frame):

    def __init__ (self, master=None, downloader=None, **options):
        super().__init__(master, **options)
        self.downloader = downloader or get_downloader()
        self.downloader.update_callback = self._update_states
        self.on_success_callback = None
        self.on_cancel_callback = None
        self._setup_widget()
    # end def

    def _on_download_cancel (self, event=None):
        self.cancel_download()
        self.update_idletasks()
        if callable(self.on_cancel_callback):
            self.after_idle(self.on_cancel_callback)
        # end if
    # end def

    def _on_download_success (self, event=None):
        self.update_idletasks()
        if callable(self.on_success_callback):
            self.after_idle(
                lambda:
                self.on_success_callback(
                    file_path=self.downloader.get_downloaded_filepath(),
                )
            )
        # end if
    # end def

    def _setup_widget (self):
        self.label_cvar = tk.StringVar()
        self.label = ttk.Label(self, textvariable=self.label_cvar)
        self.label.pack(expand=1, fill=tk.X, pady=5)
        self.progressbar_cvar = tk.DoubleVar()
        self.progressbar = ttk.Progressbar(
            self, orient=tk.HORIZONTAL, variable=self.progressbar_cvar
        )
        self.progressbar.pack(side=tk.LEFT, expand=1, fill=tk.X, pady=5)
        self.button = ttk.Button(
            self, text=_("Cancel"), command=self._on_download_cancel
        )
        self.button.pack(side=tk.RIGHT, padx=5, pady=5)
    # end def

    def _start_download (self, url=None, download_directory=None):
        self.downloader.start_download(
            url=url, download_directory=download_directory
        )
    # end def

    def _update_states (self, status="", info="", progress=0, **kw):
        self.label_cvar.set(
            "\n".join((
                str(info),
                _("URL: {url}").format(url=kw.get("shorten_url") or "")
            )).strip()
        )
        self.progressbar_cvar.set(progress)
        if status == "URL_OPEN":
            self.progressbar.configure(mode="indeterminate")
            self.progressbar.start()
        elif status == "DOWNLOAD":
            self.progressbar.configure(mode="determinate")
            self.progressbar.stop()
        elif status == "EXIT_SUCCESS":
            # better use scheduled task to release ending thread
            self.update_idletasks()
            self.after_idle(self._on_download_success)
        # end if
    # end def

    def cancel_download (self, timeout=None):
        self.update_idletasks()
        self.downloader.cancel_download(timeout)
    # end def

    def download (self, url, download_directory, on_success=None,
                                                        on_cancel=None):
        self.on_success_callback = on_success
        self.on_cancel_callback = on_cancel
        self._start_download(url, download_directory)
    # end def

    def reset (self, *args, **kw):
        self._update_states()
    # end def

# end class


def run_demo ():
    TkinterDemo().run()
# end def


if __name__ == "__main__":
    run_demo()
# end if
